import os
import posixpath
import mock
import pyxet
import secrets
import string
import pyxet
import mlflow
import pytest
from mlflow.utils.file_utils import TempDir
from mlflow.store.artifact.artifact_repository_registry import get_artifact_repository
from mlflow_xet_plugin.xet_artifact import XetHubArtifactRepository
from mlflow import log_artifact, log_artifacts, get_artifact_uri, create_experiment, MlflowClient

from mlflow.entities import (Experiment, Run, RunInfo, RunData, RunTag, Metric,
                             Param, ExperimentTag, RunStatus, LifecycleStage, FileInfo)

@pytest.fixture # run before each test function to which it is applied
def run():
    # start_mlflow_server_for_xethub()
    current_run = mlflow.active_run()
    if current_run is not None:
        with open("hello.txt", "w") as f:
            f.write("world!")
            log_artifact("hello.txt")
        yield current_run
    else:
        with mlflow.start_run() as run:
            with open("hello.txt", "w") as f:
                f.write("world!")
                log_artifact("hello.txt")
            yield run

def get_user_info():
    user = os.getenv('XET_TEST_USER')
    assert user is not None

    # host = os.getenv('XET_ENDPOINT')
    # if host is None:
    #     host = 'https://hub.xetsvc.com/'

    return {
        "user": user,
    }

# Expect a test repo whose main branch is empty (only .gitattributes)
def get_repo():
    repo = os.getenv('XET_TEST_REPO')
    assert repo is not None
    user_info = get_user_info()
    user = user_info['user']
    assert user is not None
    # set by XET_TEST_REPO
    repo_url = f"xet://{user}/{repo}"
    return repo_url


def random_string(N):
    return ''.join(secrets.choice(string.ascii_letters + string.digits)
              for i in range(N))

# Make a random branch copying src_branch in repo in format xet://[user]/[repo],
# returns the new branch name
def new_random_branch_from(repo, src_branch):
    dest_branch = random_string(20)
    pyxet.BranchCLI.make(repo, src_branch, dest_branch)
    return dest_branch

def xet_repo_mock():
    repo_url = get_repo()
    # branch = new_random_branch_from(repo_url, "main")

    return repo_url+"/main/"

repo_uri = xet_repo_mock()

def start_mlflow_server_for_xethub():
    import subprocess

    # Define the CLI command as a list of strings
    command = ["mlflow", "ui", "--backend-store-uri", "./mlruns", "--artifacts-destination", repo_uri, "--default-artifact-root", repo_uri]

    # Run the CLI command and capture the output
    try:
        result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, check=True)
        # Access the output and error (if any)
        output = result.stdout
        error = result.stderr

        # Print the output and error
        print("Output:")
        print(output)

        print("Error:")
        print(error)

        # Get the return code (0 for success)
        return_code = result.returncode
        print("Return Code:", return_code)
    except subprocess.CalledProcessError as e:
        # Handle any errors or exceptions
        print("Error:", e)

def test_get_artifact_uri(run):
    assert(mlflow.active_run())
    assert(get_artifact_uri()==run.info.artifact_uri)

def test_mlflowClient_list_artifacts_and_is_dir(run):
    artifact_uri = run.info.artifact_uri 
    
    client = MlflowClient()
    artifacts = client.list_artifacts(run.info.run_id)
    assert(artifacts)
    fs = pyxet.XetFS()
    for artifact in artifacts:
        artifact_path = posixpath.join(artifact_uri, artifact.path)
        if artifact.is_dir:
            assert(fs.isdir(artifact_path))
        else:
            assert(not fs.isdir(artifact_path))

def test_plugin_list_artifacts_and_is_dir(run):
    artifact_uri = run.info.artifact_uri 
    repository = get_artifact_repository(artifact_uri)
    artifacts = repository.list_artifacts()
    assert(artifacts)
    fs = pyxet.XetFS()
    for artifact in artifacts:
        artifact_path = posixpath.join(artifact_uri, artifact.path)
        if artifact.is_dir:
            assert(fs.isdir(artifact_path))
        else:
            assert(not fs.isdir(artifact_path))

def test_log_artifacts(run):
    artifact_uri = run.info.artifact_uri
    if not os.path.exists("outputs"):
        os.makedirs("outputs")

    if not os.path.exists("outputs/nested"):
        os.makedirs("outputs/nested")

    with open("outputs/nested/nest.txt", "w") as f:
        f.write("nested!")

    try:
        log_artifacts("outputs/nested")
        pass
    except Exception as e:
        assert(False), f"Error logging artifacts from {artifact_uri}: {e}"
    

def test_log_artifact(run):
    artifact_uri = run.info.artifact_uri
    if not os.path.exists("outputs"):
        os.makedirs("outputs")

    with open("outputs/hello.txt", "w") as f:
        f.write("world!")

    try:
        log_artifact("outputs/hello.txt")
        pass
    except Exception as e:
        assert(False), f"Error logging artifacts from {artifact_uri}: {e}"

def test_log_artifact_idempotent(run):
    test_log_artifact(run)
    test_log_artifact(run)

def test_log_artifacts_idempotent(run):
    test_log_artifacts(run)
    test_log_artifacts(run)

def test_log_and_load_text(run):
    artifact_uri = run.info.artifact_uri
    text_file = "hello.txt"
    mlflow.log_text("hello", text_file)
    file_content = mlflow.artifacts.load_text(text_file)
    assert(file_content)

def test_log_and_load_image(run):
    artifact_uri = run.info.artifact_uri
    from PIL import Image

    image = Image.new("RGB", (100, 100))
    image_file = "image.png"
    mlflow.log_image(image, image_file)
    full_path = posixpath.join(artifact_uri, image_file)
    print(f'loading image from {full_path}')
    image = mlflow.artifacts.load_image(full_path)
    assert(image)

def test_log_and_load_dict(run):
    artifact_uri = run.info.artifact_uri
    json_file = "config.json"
    mlflow.log_dict({"mlflow-version": "0.28", "n_cores": "10"}, json_file)
    full_path = posixpath.join(artifact_uri, json_file)
    config_json = mlflow.artifacts.load_dict(full_path)
    assert(config_json)

def test_log_and_load_table():
    # log and load table
    table_dict = {
        "inputs": ["What is MLflow?", "What is Databricks?"],
        "outputs": ["MLflow is ...", "Databricks is ..."],
        "toxicity": [0, "0.0"],
    }
    json_file = "table.json"
    mlflow.log_table(data=table_dict, artifact_file=json_file)
    config_json = mlflow.load_table(json_file)
    assert(config_json is not None)

def test_log_figure(run):
    artifact_uri = run.info.artifact_uri
    import matplotlib.pyplot as plt
    figure_file = "figure.png"
    full_path = posixpath.join(artifact_uri, figure_file)
    fig, ax = plt.subplots()
    ax.plot([0, 1], [2, 3])
    try:
        mlflow.log_figure(fig, figure_file)
        try:
            fs = pyxet.XetFS()
            fs.info(full_path)
        except:
            assert(False)
    except:
        assert(False)


def test_client_download_artifacts(run):
    client = MlflowClient()
    run_id = run.info.run_id
    print(client.download_artifacts(run_id, "", "./"))
    assert(client.download_artifacts(run_id, "", "./"))
    assert(client.download_artifacts(run_id, "hello.txt", "./"))

def test_client_download_artifacts_idempotent(run):
    test_client_download_artifacts(run)
    test_client_download_artifacts(run)

def test_artifact_download_artifacts(run):
    artifact_uri = run.info.artifact_uri
    print(mlflow.artifacts.download_artifacts(artifact_uri))
    assert(mlflow.artifacts.download_artifacts(artifact_uri))

def test_artifact_download_artifacts_idempotent(run):
    test_artifact_download_artifacts(run)
    test_artifact_download_artifacts(run)

def test_delete_artifacts(run):
    artifact_uri = run.info.artifact_uri
    repository = get_artifact_repository(artifact_uri)
    try:
        repository.delete_artifacts(artifact_uri)
        pass
    except Exception as e:
        assert(False), f"Error deleting artifacts from {artifact_uri}: {e}"

def test_delete_artifacts_idempotent(run):
    test_delete_artifacts(run)
    test_delete_artifacts(run)
    
@pytest.fixture
def create_nested_dirs():
    # Set up code for the fixture
    fs = pyxet.XetFS()
    user = os.getenv('XET_TEST_USER')
    repo = os.getenv('XET_TEST_REPO')
    branch = "main"
    path = "xet://" + posixpath.join(user, repo, branch, "12345/model/modelfile")
    try:
        fs.info(path)
    except:
        with fs.transaction as tr:
            commit_msg = "create test dir"
            tr.set_commit_message(commit_msg)
            file = fs.open(path, 'w')
            file.write('')
            file.close()
    yield
    # Teardown code for the fixture (if needed)

@pytest.mark.parametrize("base_uri, download_arg, list_return_val", [
    ('12345/model', '', ['modelfile']),
    ('12345/model', '', ['.', 'modelfile']),
    ('12345', 'model', ['model/modelfile']),
    ('12345', 'model', ['model', 'model/modelfile']),
    ('', '12345/model', ['12345/model/modelfile']),
    ('', '12345/model', ['12345/model', '12345/model/modelfile']),
])
def test_download_artifacts_does_not_infinitely_loop(base_uri, download_arg, list_return_val, create_nested_dirs):
    base_uri = repo_uri + base_uri
    def list_artifacts(path):
        fullpath = posixpath.join(base_uri, path)
        if fullpath.endswith("model") or fullpath.endswith("model/"):
            return [FileInfo(item, False, 123) for item in list_return_val]
        elif fullpath.endswith("12345") or fullpath.endswith("12345/"):
            return [FileInfo(posixpath.join(path, "model"), True, 0)]
        else:
            return []

    with mock.patch.object(XetHubArtifactRepository, "list_artifacts") as list_artifacts_mock:
        list_artifacts_mock.side_effect = list_artifacts
        repo = XetHubArtifactRepository(base_uri)
        repo.download_artifacts(download_arg)


@pytest.mark.parametrize("base_uri, download_arg, list_return_val", [
    ('', '12345/model', ['12345/model', '12345/model/modelfile', '12345/model/emptydir']),
])
def test_download_artifacts_handles_empty_dir(base_uri, download_arg, list_return_val):
    base_uri = repo_uri + base_uri
    def list_artifacts(path):
        if path.endswith("model"):
            return [FileInfo(item, item.endswith("emptydir"), 123) for item in list_return_val]
        elif path.endswith("12345") or path.endswith("12345/"):
            return [FileInfo("12345/model", True, 0)]
        else:
            return []

    with mock.patch.object(XetHubArtifactRepository, "list_artifacts") as list_artifacts_mock:
        list_artifacts_mock.side_effect = list_artifacts
        repo = XetHubArtifactRepository(base_uri)
        with TempDir() as tmp:
            repo.download_artifacts(download_arg, dst_path=tmp.path())



import os
import sys
import pyxet
import posixpath
from mlflow.exceptions import MlflowException
from mlflow.entities import FileInfo
from mlflow.utils.file_utils import relative_path_to_artifact_path
from mlflow.store.artifact.artifact_repo import ArtifactRepository


class XetHubArtifactRepository(ArtifactRepository):
    """Stores artifacts on XetHub."""

    # artifact_uri indicates where all artifacts for a mlflow run are stored, 
    # should be in the format of xet://[user]/[repo]/[branch]/[experiment_id]/[run_id]/artifacts
    # e.g. xet://keltonzhang/mlflowArtifacts/main/0/7574037f153b4cd7819453e8fb466ae9/artifacts
    def __init__(self, artifact_uri, xet_client=None):

        super(XetHubArtifactRepository, self).__init__(artifact_uri)

        # Allow override for testing
        if xet_client:
            self.xet_client = xet_client
            return
        
        self.xet_client = pyxet
        print(f"Artifacts located at {self.artifact_uri}")
        # strip trailing slash as posix join will add slash in between paths
        if artifact_uri.endswith("/"):
            self.artifact_uri = artifact_uri[:-1]

        # pathComponents = self.artifact_uri.split("/")
        # if len(pathComponents) < 8:
        #     raise Exception("Invalid artifact URI format, check if the artifact destination/root passed to your MLflow server is of the form xet://user/repo/branch")
    

    """
        Log a local file as an artifact, optionally taking an ``artifact_path`` to place it in
        within the run's artifacts. Run artifacts can be organized into directories, so you can
        place the artifact in a directory this way.

        :param local_file: Path to artifact to log
        :param artifact_path: Directory within the run's artifact directory in which to log the
                              artifact.
    """
    def log_artifact(self, local_file, artifact_path=None):

        # dest path would be formatted as xet://user/repo/branch/mlflow_experiment_group/mlflow_run_id/artifacts/file
        if artifact_path:
            dest_path = posixpath.join(self.artifact_uri, artifact_path)
        else:
            dest_path = posixpath.join(self.artifact_uri, os.path.basename(local_file))
            
        # Store file to XetHub
        fs = self.xet_client.XetFS()
        commit_msg = "Log artifact %s" % os.path.basename(local_file)

        sys.stdout.write(f"Logging artifact to XetHub from {local_file} to {dest_path}\n")
        with fs.transaction as tr:
            tr.set_commit_message(commit_msg)
            dest_file = fs.open(dest_path, 'wb')
            src_file = open(local_file, 'rb').read()
            dest_file.write(src_file)
            dest_file.close()

        sys.stdout.write(f"Logged artifact to XetHub from {local_file} to {dest_path}\n")

    """
        Log the files in the specified local directory as artifacts, optionally taking
        an ``artifact_path`` to place them in within the run's artifacts.

        :param local_dir: Directory of local artifacts to log
        :param artifact_path: Directory within the run's artifact directory in which to log the
                              artifacts
    """
    def log_artifacts(self, local_dir, artifact_path=None):
        # remote, branch, path = self.xet_client.parse_url(self.artifact_uri)
        dest_path = self.artifact_uri
        if artifact_path:
            dest_path = posixpath.join(dest_path, artifact_path)

        fs = self.xet_client.XetFS()
        local_dir = os.path.abspath(local_dir)
        commit_msg = "Log artifacts under %s" % os.path.basename(local_dir)

        sys.stdout.write(f"Logging artifacts to XetHub from {local_dir} to {dest_path}\n")
        with fs.transaction as tr:
            tr.set_commit_message(commit_msg)
            for (root, _, filenames) in os.walk(local_dir):
                upload_path = dest_path
                if root != local_dir:
                    rel_path = os.path.relpath(root, local_dir)
                    rel_path = relative_path_to_artifact_path(rel_path)
                    upload_path = posixpath.join(dest_path, rel_path)
                for f in filenames:
                    # self.log_artifact(os.path.join(root, f), os.path.join(upload_path, f))
                    local_file = os.path.join(root, f)
                    file_dest_path = os.path.join(upload_path, f)
                    dest_file = fs.open(file_dest_path, 'wb')
                    src_file = open(local_file, 'rb').read()
                    dest_file.write(src_file)
                    dest_file.close()
                
        sys.stdout.write(f"Logged artifacts to XetHub from {local_dir} to {dest_path}\n")

    """
        Return all the artifacts for this run_id directly under path. If path is a file, returns
        an empty list. Will error if path is neither a file nor directory.

        :param path: Relative source path that contains desired artifacts

        :return: List of artifacts as FileInfo listed directly under path.
    """
    def list_artifacts(self, path=None):
        artifact_path = self.artifact_uri
        
        dest_path = artifact_path
        if path:
            dest_path = posixpath.join(dest_path, path)

        print("Listing artifacts of %s\n" % dest_path)

        infos = []
        dest_path = dest_path + "/" if dest_path else ""
        fs = self.xet_client.XetFS()
        if fs.isdir(dest_path):
            entries = fs.ls(dest_path)

            for entry in entries:
                entryName = entry["name"]
                entryType = entry["type"]
                entrySize = entry["size"]
                self._verify_listed_entry_contains_artifact_path_prefix(
                        listed_entry_path="xet://"+entryName, artifact_path=artifact_path)
                
                start_path = artifact_path[6:] # remove xet:// from xet://[user]/[repo]/[branch]/[experiment_id]/[run_id]/artifacts
                if entryType=="file":
                    # is file
                    file_path = entryName
                    file_rel_path = posixpath.relpath(path=file_path, start=start_path)
                    file_size = entrySize
                    infos.append(FileInfo(file_rel_path, False, file_size))
                else:
                    # is dir
                    subdir_path = entryName
                    subdir_rel_path = posixpath.relpath(path=subdir_path, start=start_path)
                    infos.append(FileInfo(subdir_rel_path, True, None))

        else:
            # the path is a single file
            pass

        print(f"Listed artifacts: {infos}")
        return sorted(infos, key=lambda f: f.path)

    @staticmethod
    def _verify_listed_entry_contains_artifact_path_prefix(listed_entry_path, artifact_path):
        if not listed_entry_path.startswith(artifact_path):
            raise MlflowException(
                "The path of the listed xet entry does not begin with the specified"
                " artifact path. Artifact path: {artifact_path}. entry path:"
                " {entry_path}.".format(
                    artifact_path=artifact_path, entry_path=listed_entry_path))

    def download_artifacts(self, artifact_path, dst_path=None):
        """
        Artifacts tracked by the plugin already exist on the local filesystem.
        If ``dst_path`` is ``None``, the absolute filesystem path of the specified artifact is
        returned. If ``dst_path`` is not ``None``, the local artifact is copied to ``dst_path``.

        :param artifact_path: Relative source path to the desired artifacts.
        :param dst_path: Absolute path of the local filesystem destination directory to which to
                         download the specified artifacts. This directory must already exist. If
                         unspecified, the absolute path of the local artifact will be returned.

        :return: Absolute path of the local filesystem location containing the desired artifacts.
        """
        print(f"dst_path {dst_path}")
        if dst_path:
            return super().download_artifacts(artifact_path, dst_path)
        # NOTE: The artifact_path is expected to be in posix format.
        # Posix paths work fine on windows but just in case we normalize it here.
        # local_artifact_path = os.path.join(self.artifact_dir, os.path.normpath(artifact_path))
        # if not os.path.exists(local_artifact_path):
        #     raise OSError(f"No such file or directory: '{local_artifact_path}'")        
        else:
            rel_artifact_path = artifact_path
            artifact_path = posixpath.join(self.artifact_uri, artifact_path)

            # artifact_path is of the format xet://user/repo/branch/mlflow_experiment_group/mlflow_run_id/artifacts/file
            mlflow_subpath = "/".join(artifact_path.split("/")[5:])
            dst_path = os.path.abspath("./mlruns/"+mlflow_subpath)

            fs = self.xet_client.XetFS()
            if fs.isdir(artifact_path):
                print(f"Downloading artifacts from {artifact_path} to {dst_path}\n")
                fs.get(artifact_path, dst_path, recursive=True)
                print(f"Downloaded artifacts from {artifact_path} to {dst_path}")
            else:
                self._download_file(rel_artifact_path, dst_path)

            return dst_path

    def _download_file(self, remote_file_path, local_path):
        fs = self.xet_client.XetFS()
        xet_root_path = self.artifact_uri
        xet_full_path = posixpath.join(xet_root_path, remote_file_path)
        print(f"Downloading artifact from {xet_full_path} to {local_path}\n")
        fs.get(xet_full_path, local_path)
        print(f"Downloaded artifact from {xet_full_path} to {local_path}\n")

    def delete_artifacts(self, artifact_path=None):
        fs = self.xet_client.XetFS()
        if fs.isdir(artifact_path):
            commit_msg = "Delete artifacts in %s" % os.path.basename(artifact_path)
            print("Deleting artifacts from %s\n" % (artifact_path))
            with fs.transaction as tr:
                tr.set_commit_message(commit_msg)
                cli = self.xet_client.PyxetCLI()
                cli.rm(artifact_path)
                # for entry in self.list_artifacts(artifact_path):
                #     fs.rm(entry)
            print("Deleted artifacts from %s\n" % (artifact_path))
        else:
            commit_msg = "Delete artifact %s" % os.path.basename(artifact_path)
            print("Deleting artifact %s\n" % (artifact_path))
            with fs.transaction as tr:
                tr.set_commit_message(commit_msg)
                fs.rm(artifact_path)
            print("Deleted artifact %s\n" % (artifact_path))

        


"""
To run a tracking server with this app, use `mlflow server --app-name custom_app`.
"""
import logging

# This would be all that plugin author is required to import
from mlflow.server import app as custom_app

# Can do custom logging on either the app or logging itself
# but you'll possibly have to clear the existing handlers or there will be duplicate output
# See https://docs.python.org/3/howto/logging-cookbook.html

app_logger = logging.getLogger(__name__)

# Configure the app
custom_app.config["MY_VAR"] = "config-var"
app_logger.warning(f"Using {__name__}")


def is_logged_in():
    return True


@custom_app.before_request
def before_req_hook():
    """A custom before request handler.

    Can implement things such as authentication, special handling, etc.
    """
    if not is_logged_in():
        app_logger.warning("Hello from before request!")
        return "Unauthorized", 403


@custom_app.route("/custom/endpoint", methods=["GET"])
def custom_endpoint():
    """A custom endpoint."""
    return "custom_endpoint", 200



from setuptools import setup, find_packages

setup(
    name = "mlflow-xethub",
    version = '0.0.0',
    url = "https://github.com/xetdata/Xet-MLflow",
    description = "mlflow[xethub] is a mlflow plugin integrating XetHub with MLflow so that you can use existing MLflow code to track experiments but store artifacts to XetHub.",
    long_description = open('README.md').read(),
    long_description_content_type = "text/markdown",
    author = "XetHub",
    author_email = "contact@xethub.com",
    maintainer = "Kelton Zhang",
    maintainer_email = "kelton@xethub.com",
    license = "BSD-3-Clause",
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
    ],
    keywords = ["ai",
        "collaboration",
        "data-science",
        "developer-tools",
        "git",
        "mlflow",
        "machine-learning",
        "reproducibility",
    ],
    packages=find_packages(),
    # Require MLflow as a dependency of the plugin, so that plugin users can simply install
    # the plugin and then immediately use it with MLflow
    install_requires=["mlflow>=1.27.0", "pyxet>=0.1.0"],
    python_requires='>=3.7',
    entry_points={
        "mlflow.artifact_repository": "xet=mlflow_xet_plugin.xet_artifact:XetHubArtifactRepository",
    },
)



import urllib.parse

from mlflow.store.tracking.file_store import FileStore


class PluginFileStore(FileStore):
    """FileStore provided through entrypoints system"""

    def __init__(self, store_uri=None, artifact_uri=None):
        path = urllib.parse.urlparse(store_uri).path if store_uri else None
        self.is_plugin = True
        super().__init__(path, artifact_uri)
